import numpy as np

class EqualityConstrainedProblem:
    def __init__(self, f, df, hess, A, b):
        self.f = f
        self.df = df
        self.hess = hess
        self.A = A
        self.b = b
        self.p = A.shape[0]
        self.n = A.shape[1]

    def newton_step(self, x, nu=None):
        # nu == None is used for Newton’s method with equality constraints
        # nu != None is used for Infeasible start Newton method
        n = self.n
        p = self.p

        AugMatrix = np.zeros((n+p,n+p))
        AugMatrix[:n,:n] = self.hess(x)
        AugMatrix[:n,n:] = self.A.T
        AugMatrix[n:,:n] = self.A

        augvec = np.zeros(n+p)
        augvec[:n] = -self.df(x)
        augvec[n:] = self.b - self.A @ x
        soln = np.linalg.solve(AugMatrix, augvec)

        if nu is not None:
            return soln[:n], soln[n:] - nu
        else:
            return soln[:n]

    def backtracking_line_search(self, x, d, alpha, beta):
        k = 0
        t = 1.0
        while self.f(x + t * d) > self.f(x) + t * alpha * np.dot(self.df(x), d):
            t *= beta
            k += 1
        return t, k

    def newton_with_linear_constraint(self, x, alpha, beta, epsilon, maxiter=100):

        x_list = [x]
        for _ in range(maxiter):
            dx = self.newton_step(x)
            if np.dot(-self.df(x), dx) < epsilon:
                break
            else:
                # line search
                t, _ = self.backtracking_line_search(x, dx, alpha, beta)
                x = x + t * dx
                x_list.append(x)
        return x, x_list

    def r(self, x, nu):
        r = np.linalg.norm(self.df(x) + self.A.T @ nu)**2 + np.linalg.norm(self.A @ x - self.b)**2
        return np.sqrt(r)

    def newton_with_infeasible_start(self, x, nu, alpha, beta, epsilon,
                                     maxiter=100, max_line_search=100):

        x_list = [x]
        nu_list = [nu]

        for _ in range(maxiter):
            dx, dnu = self.newton_step(x, nu=nu)
            r0 = self.r(x, nu)

            if r0  < epsilon:
                break

            # line search
            t = 1
            count = 0
            while True:
                xt = x + t * dx
                nut = nu + t * dnu
                r = self.r(xt, nut)

                if r > (1-alpha * t) * r0:
                    t *= beta
                    count += 1
                else:
                    break

                if count > max_line_search:
                    print("Exceed max line search")
                    break

            x = x + t * dx
            nu = nu + t * dnu
            x_list.append(x)
            nu_list.append(nu)

        return x, x_list, nu, nu_list
